@@ -10,15 +10,13 @@ from trytond.model import ModelView, ModelSQL, MatchMixin, fields, \
 from trytond.pyson import Eval, If
 from trytond.pool import Pool, PoolMeta
 from trytond.transaction import Transaction
-from trytond import backend
 
 from trytond.modules.product import price_digits
 
 __all__ = ['Template', 'Product', 'ProductSupplier', 'ProductSupplierPrice']
 
 
-class Template:
-    __metaclass__ = PoolMeta
+class Template(metaclass=PoolMeta):
     __name__ = "product.template"
     purchasable = fields.Boolean('Purchasable', states={
             'readonly': ~Eval('active', True),
@@ -44,16 +42,6 @@ class Template:
                 'change_purchase_uom': ('Purchase prices are based '
                     'on the purchase uom.'),
                 })
-        required = ~Eval('account_category') & Eval('purchasable', False)
-        if not cls.account_expense.states.get('required'):
-            cls.account_expense.states['required'] = required
-        else:
-            cls.account_expense.states['required'] = (
-                cls.account_expense.states['required'] | required)
-        if 'account_category' not in cls.account_expense.depends:
-            cls.account_expense.depends.append('account_category')
-        if 'purchasable' not in cls.account_expense.depends:
-            cls.account_expense.depends.append('purchasable')
 
     @fields.depends('default_uom', 'purchase_uom', 'purchasable')
     def on_change_default_uom(self):
@@ -95,8 +83,7 @@ class Template:
         super(Template, cls).write(*args)
 
 
-class Product:
-    __metaclass__ = PoolMeta
+class Product(metaclass=PoolMeta):
     __name__ = 'product.product'
 
     purchase_price_uom = fields.Function(fields.Numeric(
@@ -128,6 +115,8 @@ class Product:
         context = Transaction().context
         prices = {}
 
+        assert len(products) == len(set(products)), "Duplicate products"
+
         uom = None
         if context.get('uom'):
             uom = Uom(context['uom'])
@@ -148,7 +137,13 @@ class Product:
             else:
                 product_uom = uom
             pattern = ProductSupplier.get_pattern()
-            for product_supplier in product.product_suppliers:
+            product_suppliers = product.product_suppliers
+            if context.get('product_supplier'):
+                for product_supplier in product.product_suppliers:
+                    if product_supplier.id == context['product_supplier']:
+                        product_suppliers = [product_supplier]
+                        break
+            for product_supplier in product_suppliers:
                 if product_supplier.match(pattern):
                     pattern = ProductSupplierPrice.get_pattern()
                     for price in product_supplier.prices:
@@ -190,42 +185,13 @@ class ProductSupplier(sequence_ordered(), ModelSQL, ModelView, MatchMixin):
 
     @classmethod
     def __register__(cls, module_name):
-        TableHandler = backend.get('TableHandler')
         transaction = Transaction()
         cursor = transaction.connection.cursor()
-        table = TableHandler(cls, module_name)
+        table = cls.__table_handler__(module_name)
         sql_table = cls.__table__()
 
-        # Migration from 2.2 new field currency
-        created_currency = table.column_exist('currency')
-
         super(ProductSupplier, cls).__register__(module_name)
 
-        # Migration from 2.2 fill currency
-        if not created_currency:
-            Company = Pool().get('company.company')
-            company = Company.__table__()
-            limit = transaction.database.IN_MAX
-            cursor.execute(*sql_table.select(Count(sql_table.id)))
-            product_supplier_count, = cursor.fetchone()
-            for offset in range(0, product_supplier_count, limit):
-                cursor.execute(*sql_table.join(company,
-                        condition=sql_table.company == company.id
-                        ).select(sql_table.id, company.currency,
-                        order_by=sql_table.id,
-                        limit=limit, offset=offset))
-                for product_supplier_id, currency_id in cursor.fetchall():
-                    cursor.execute(*sql_table.update(
-                            columns=[sql_table.currency],
-                            values=[currency_id],
-                            where=sql_table.id == product_supplier_id))
-
-        # Migration from 2.4: drop required on sequence
-        table.not_null_action('sequence', action='remove')
-
-        # Migration from 2.6: drop required on delivery_time
-        table.not_null_action('delivery_time', action='remove')
-
         # Migration from 3.8: change delivery_time inte timedelta lead_time
         if table.column_exist('delivery_time'):
             cursor.execute(*sql_table.select(
@@ -266,7 +232,11 @@ class ProductSupplier(sequence_ordered(), ModelSQL, ModelView, MatchMixin):
                 self.currency, = row
 
     def get_rec_name(self, name):
-        return '%s @ %s' % (self.product.rec_name, self.party.rec_name)
+        if self.code:
+            name = self.name or self.product.name
+            return '[' + self.code + '] ' + name
+        else:
+            return self.name or self.product.rec_name
 
     @classmethod
     def search_rec_name(cls, name, clause):
@@ -274,10 +244,13 @@ class ProductSupplier(sequence_ordered(), ModelSQL, ModelView, MatchMixin):
             bool_op = 'AND'
         else:
             bool_op = 'OR'
-        return [bool_op,
+        domain = [bool_op,
             ('product',) + tuple(clause[1:]),
             ('party',) + tuple(clause[1:]),
+            ('code',) + tuple(clause[1:]),
+            ('name',) + tuple(clause[1:]),
             ]
+        return domain
 
     @property
     def uom(self):
@@ -319,15 +292,18 @@ class ProductSupplierPrice(
     __name__ = 'purchase.product_supplier.price'
     product_supplier = fields.Many2One('purchase.product_supplier',
             'Supplier', required=True, ondelete='CASCADE')
-    quantity = fields.Float('Quantity', required=True, help='Minimal quantity')
+    quantity = fields.Float(
+        "Quantity",
+        required=True,
+        domain=[('quantity', '>=', 0)],
+        help='Minimal quantity')
     unit_price = fields.Numeric('Unit Price', required=True,
         digits=price_digits)
 
     @classmethod
     def __register__(cls, module_name):
-        TableHandler = backend.get('TableHandler')
         cursor = Transaction().connection.cursor()
-        table = TableHandler(cls, module_name)
+        table = cls.__table_handler__(module_name)
         sql_table = cls.__table__()
 
         fill_sequence = not table.column_exist('sequence')
@@ -352,6 +328,6 @@ class ProductSupplierPrice(
         Uom = pool.get('product.uom')
         test_quantity = Uom.compute_qty(
             self.product_supplier.uom, self.quantity, uom)
-        if test_quantity > quantity:
+        if test_quantity > abs(quantity):
             return False
         return super(ProductSupplierPrice, self).match(pattern)
